import datetime
import time
import json
from uuid import UUID
import re
from urllib.parse import quote_plus


def camelcase_to_dash(s):
    # stolen from http://stackoverflow.com/a/1176023
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1-\2', s)
    return re.sub('([a-z0-9])([A-Z])', r'\1-\2', s1).lower()


class RaiseNotImplemented(object):
    def __get__(self, instance, owner):
        raise NotImplementedError

RAISE_NOT_IMPLEMENTED = RaiseNotImplemented()


def convert_datetime_to_str(dt):
    return dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def type_handler(x):
    if isinstance(x, datetime.datetime):
        return x.strftime("%Y-%m-%dT%H:%M:%SZ")

    if isinstance(x, UUID):
        return str(x)

    raise TypeError("Unknown type")


def json_dump(*args, **kwargs):
    kwargs.setdefault('default', type_handler)

    return json.dumps(*args, **kwargs)


def build_url_path(parts):
    return '/'.join([quote_plus(x) for x in parts])


def datetime_to_secs(dt):
    return int(time.mktime(dt.timetuple()))



"""
Collection Managers
-------------------

Collection managers brings together the item manager,
collections and storages.

"""
import six
from .utils import RAISE_NOT_IMPLEMENTED
from .collections import Collection


class CManagerMetaclass(type):
    """
    This metaclass sets a dictionary named `_declared_collections` on the class.
    Any instances of `Collection` included as attributes on either the class
    or on any of its superclasses will be include in the
    `_declared_collections` dictionary.
    """

    @classmethod
    def _get_declared_collections(cls, bases, attrs):
        fields = [(collection_name, attrs.get(collection_name))
                  for collection_name, obj in list(attrs.items())
                  if isinstance(obj, Collection)]

        # If this class is subclassing another Serializer, add that Serializer's
        # fields.  Note that we loop over the bases in *reverse*. This is necessary
        # in order to maintain the correct order of fields.
        for base in reversed(bases):
            if hasattr(base, '_declared_collections'):
                fields = [
                    (collection_name, obj) for collection_name, obj
                    in base._declared_collections.items()
                    if collection_name not in attrs
                ] + fields

        return dict(fields)

    def __new__(cls, name, bases, attrs):
        attrs['_declared_collections'] = cls._get_declared_collections(bases, attrs)
        return super(CManagerMetaclass, cls).__new__(cls, name, bases, attrs)


@six.add_metaclass(CManagerMetaclass)
class CManager(object):
    item_manager = RAISE_NOT_IMPLEMENTED

    def __init__(self, storage):
        self.storage = storage

        for collection in self._declared_collections.values():
            collection.set_storage(self.storage)
            self.item_manager.set_storage(self.storage)
            collection.set_item_manager(self.item_manager)

    async def add_item(self, item):
        await self.item_manager.add_item(item)
        awaitable_resps = {}
        for name, collection in self._declared_collections.items():
            awaitable_resps[name] = collection.add_item(item)

        resps = {}
        for name, resp in awaitable_resps.items():
            resps[name] = await resp

        return resps

    async def remove_item(self, item):
        awaitable_resps = {}
        for name, collection in self._declared_collections.items():
            awaitable_resps[name] = collection.remove_item(item)

        resps = {}
        for name, resp in awaitable_resps.items():
            resps[name] = await resp

        await self.item_manager.remove_item(item)

        return resps



import pygtrie


class InMemoryStorage(object):

    def __init__(self):
        self.t = pygtrie.CharTrie()

    async def create_key(self, storage_key, data=None):
        data = data if data else bytes()

        self.t[storage_key] = data

    async def remove_key(self, storage_key):
        del self.t[storage_key]

    async def remove_keys(self, storage_keys):

        for key in storage_keys:
            del self.t[key]

    async def get_key(self, storage_key):

        return {
            'data': self.t.get(storage_key),
            'metadata': {},
        }

    def _build_keys(self, key_iterator, max_keys, continuation_token):
        keys = []
        found_token = False

        for x in key_iterator:
            if continuation_token and found_token is False:
                if x != continuation_token:
                    continue

                found_token = True

            keys.append(x)

            if len(keys) == max_keys:
                break

        return keys

    async def get_keys(self, prefix, max_keys=200, continuation_token=None):

        key_iterator = self.t.iterkeys(prefix)

        try:
            keys = self._build_keys(key_iterator, max_keys, continuation_token)
        except KeyError:
            return {
                'keys': []
            }

        try:
            continuation_token = next(key_iterator)
        except StopIteration:
            continuation_token = None

        ret = {
            'keys': keys,
        }

        if continuation_token:
            ret['continuation_token'] = continuation_token

        return ret



"""
Item Managers
-------------

An item manager allows you persist items, and retrieve
them by keys.

"""

from .utils import camelcase_to_dash, build_url_path


class IManager(object):
    def __init__(self, prefix='items', name=None, storage=None):
        self.name = name if name else camelcase_to_dash(self.__class__.__name__)
        self.prefix = prefix
        self.storage = storage

    def set_storage(self, storage):
        self.storage = storage

    def base_key_parts(self):
        return [self.prefix, self.name]

    def get_pk(self, item):
        return str(item.pk)

    def generate_key_for_pk(self, pk):

        key_parts = self.base_key_parts()
        key_parts += [pk]

        return build_url_path(key_parts)

    def serialize_item(self, item):
        raise NotImplementedError

    def deserialize_item(self, data):
        raise NotImplementedError

    async def add_item(self, item):
        key = self.generate_key_for_pk(self.get_pk(item))
        return await self.storage.create_key(key, data=self.serialize_item(item))

    async def remove_item(self, item):
        key = self.generate_key_for_pk(self.get_pk(item))
        return await self.storage.remove_key(key)

    async def get_item(self, pk):
        key = self.generate_key_for_pk(pk)
        resp = await self.storage.get_key(key)

        return self.deserialize_item(resp['data'])



import string

ALPHABET = string.ascii_uppercase + string.ascii_lowercase
ALPHABET_REVERSE = dict((c, i) for (i, c) in enumerate(ALPHABET))
BASE = len(ALPHABET)


def encode_int_as_str(n):
    s = []

    while True:
        n, r = divmod(n, BASE)
        s.append(ALPHABET[r])
        if n == 0:
            break

    return ''.join(reversed(s))


def decode_str_as_int(s):
    n = 0

    for c in s:
        n = n * BASE + ALPHABET_REVERSE[c]

    return n


def reverse(s):
    n = []
    for i in s:
        ind = ALPHABET.index(i)
        new_ind = len(ALPHABET) - ind - 1
        n += [ALPHABET[new_ind]]

    return ''.join(n)



"""
Collections
-----------

Collections allow you to put items into
queryable ordered lists.

"""

from .utils import RAISE_NOT_IMPLEMENTED, build_url_path, camelcase_to_dash


class Collection(object):

    key = RAISE_NOT_IMPLEMENTED

    def __init__(self, storage=None, prefix='collections', name=None, item_manager=None):
        self.prefix = prefix
        self.storage = storage
        self.name = name if name else camelcase_to_dash(self.__class__.__name__)
        self.item_manager = item_manager

    def set_storage(self, storage):
        self.storage = storage

    def set_item_manager(self, item_manager):
        self.item_manager = item_manager

    def get_pk(self, item):
        return str(item.pk)

    def parse_pk_from_key(self, key):
        return key.split('/')[-1]

    def base_key_parts(self):
        return [self.prefix, self.name]

    def generate_keys_for_item(self, item):
        pk = self.get_pk(item)

        for key in self.key(item):
            key_parts = self.base_key_parts()
            key_parts += key
            key_parts += [pk]

            yield build_url_path(key_parts)

    async def add_item(self, item):
        keys = []
        reqs = []
        for storage_key in self.generate_keys_for_item(item):
            keys += [storage_key]
            reqs += [self.storage.create_key(storage_key)]

        resps = [await x for x in reqs]

        return keys

    async def remove_item(self, item):
        keys = []
        reqs = []
        for storage_key in self.generate_keys_for_item(item):
            keys += [storage_key]
            reqs += [self.storage.remove_key(storage_key)]

        resps = [await x for x in reqs]

        return keys

    async def get_keys(self, max_keys=200, continuation_token=None, prefix_parts=None):
        prefix_parts = prefix_parts if prefix_parts else []

        resp = await self.storage.get_keys(
            prefix=build_url_path(self.base_key_parts() + prefix_parts),
            max_keys=max_keys,
            continuation_token=continuation_token
        )

        resp['keys'] = [self.parse_pk_from_key(x) for x in resp['keys']]

        return resp

    async def get_items(self, max_keys=200, continuation_token=None, prefix_parts=None):

        resp = await self.get_keys(
            max_keys=max_keys,
            continuation_token=continuation_token,
            prefix_parts=prefix_parts,
        )

        reqs = [self.item_manager.get_item(x) for x in resp['keys']]

        items = [await req for req in reqs]

        ret = {
            'items': items,
        }

        if 'continuation_token' in resp:
            ret['continuation_token'] = resp['continuation_token']

        return ret



#!/usr/bin/env python

import os
import re
import sys

from codecs import open

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist')
    os.system('twine upload dist/*')
    sys.exit()

packages = [
    'ssshelf',
    'ssshelf.storages'
]


requires = [
    "aiobotocore>=0.4.5",
    "paco>=0.1.11",
    "retrying>=1.3.3",
    "pygtrie>=2.2",
    "six>=1.10.0",
]

test_requirements = [
    "moto",
    "pytest",
    "requests",
    "flask",
    "attrs",
    "iso8601==0.1.12",
    "pytest-asyncio",
    "pystrainer==1.3.1",
    "simpleflake",
]

with open('ssshelf/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

with open('README.rst', 'r', 'utf-8') as f:
    readme = f.read()


setup(
    name='ssshelf',
    version=version,
    description='A collection and persistance manager for blob stores like aws S3',
    long_description=readme,
    author='Alex Kessinger',
    author_email='voidfiles@gmail.com',
    url='http://github.com/voidfiles/ssshelf',
    packages=packages,
    package_data={
        '': [
            'LICENSE',
            '*.txt',
        ],
        'ssshelf': []
    },
    package_dir={
        'ssshelf': 'ssshelf',
    },
    include_package_data=True,
    install_requires=requires,
    license='Apache 2.0',
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Development Status :: 5 - Production/Stable',
        'Topic :: Software Development',
    ],
    cmdclass={'test': PyTest},
    tests_require=test_requirements,
)



import asyncio
import aiobotocore
from retrying import retry


@retry(stop_max_attempt_number=2)
def retry_client(method, *args, **kwargs):
    return method(*args, **kwargs)


class S3Storage(object):

    def __init__(self, bucket, aws_client_kwargs=None, s3_client=None):
        self.aws_client_kwargs = aws_client_kwargs if aws_client_kwargs else {
            'region_name': 'us-east-1',
        }
        self.bucket = bucket
        self.s3_client = s3_client

    def get_s3_client(self):
        if self.s3_client:
            return self.s3_client

        session = aiobotocore.get_session(loop=loop)
        self.s3_client = session.create_client('s3', **self.aws_client_kwargs)

        return self.s3_client

    async def create_key(self, storage_key, data=None):
        data = data if data else bytes()

        await retry_client(
            self.get_s3_client().put_object,
            Bucket=self.bucket,
            Key=storage_key,
            Body=data
        )

    async def remove_key(self, storage_key):

        await retry_client(
            self.get_s3_client().delete_object,
            Bucket=self.bucket,
            Key=storage_key,
        )

    async def remove_keys(self, storage_keys):

        await retry_client(
            self.get_s3_client().delete_objects,
            Bucket=self.bucket,
            Delete={
                "Objects": [{'Key': x} for x in storage_keys],
                "Quiet": True,
            },
        )

    async def get_key(self, storage_key):
        resp = await retry_client(
            self.get_s3_client().get_object,
            Bucket=self.bucket,
            Key=storage_key,
        )

        return {
            'data': await resp.get('Body').read(),
            'metadata': resp.get('Metadata', {}),
        }

    async def get_keys(self, prefix, max_keys=200, continuation_token=None):

        resp = None

        kwargs = {
            "Bucket": self.bucket,
            "MaxKeys": max_keys,
            "Prefix": prefix,
        }

        if continuation_token:
            kwargs['ContinuationToken'] = continuation_token

        resp = await retry_client(
            self.get_s3_client().list_objects_v2,
            **kwargs,
        )

        response_keys = [x['Key'] for x in resp.get('Contents', [])]

        resp_data = {
            'keys': response_keys,
        }

        continuation_token = resp.get('NextContinuationToken')
        if continuation_token:
            resp_data['continuation_token'] = continuation_token

        return resp_data
